#include <gtest/gtest.h>
#include <platform.h>

#include "util.h"

#define TEST_COMPILE(src) DO(test_compile(src, "ir"))

#define TEST_ERRORS_1(src, err1) \
  { const char* errs[] = {err1, NULL}; \
    DO(test_expected_errors(src, "ir", errs)); }


class CodegenFFITest : public PassTest
{};


TEST_F(CodegenFFITest, ReferenceReferenceCompatibility)
{
  const char* pkg =
    "use @foo[None](x: Foo)\n"

    "primitive Foo\n"
    "  fun apply() => @foo(this)";

  const char* src =
    "use \"pkg\"\n"
    "use @foo[None](x: Main)\n"

    "actor Main\n"
    "  new create(env: Env) =>\n"
    "    @foo(this)\n"
    "    Foo()";

  add_package("pkg", pkg);

  TEST_COMPILE(src);
}


TEST_F(CodegenFFITest, ReferencePointerCompatibility)
{
  const char* pkg =
    "use @foo[None](x: Foo)\n"

    "primitive Foo\n"
    "  fun apply() => @foo(this)";

  const char* src =
    "use \"pkg\"\n"
    "use @foo[None](x: Pointer[Main])\n"

    "actor Main\n"
    "  new create(env: Env) =>\n"
    "    @foo(Pointer[Main])\n"
    "    Foo()";

  add_package("pkg", pkg);

  TEST_COMPILE(src);
}


TEST_F(CodegenFFITest, ReferenceUSizeCompatibility)
{
  const char* pkg =
    "use @foo[None](x: Foo)\n"

    "primitive Foo\n"
    "  fun apply() => @foo(this)";

  const char* src =
    "use \"pkg\"\n"
    "use @foo[None](x: USize)\n"

    "actor Main\n"
    "  new create(env: Env) =>\n"
    "    @foo(0)\n"
    "    Foo()";

  add_package("pkg", pkg);

  TEST_COMPILE(src);
}


TEST_F(CodegenFFITest, PointerPointerCompatibility)
{
  const char* pkg =
    "use @foo[None](x: Pointer[Foo])\n"

    "primitive Foo\n"
    "  fun apply() => @foo(Pointer[Foo])";

  const char* src =
    "use \"pkg\"\n"
    "use @foo[None](x: Pointer[Main])\n"

    "actor Main\n"
    "  new create(env: Env) =>\n"
    "    @foo(Pointer[Main])\n"
    "    Foo()";

  add_package("pkg", pkg);

  TEST_COMPILE(src);
}


TEST_F(CodegenFFITest, PointerUSizeCompatibility)
{
  const char* pkg =
    "use @foo[None](x: Pointer[Foo])\n"

    "primitive Foo\n"
    "  fun apply() => @foo(Pointer[Foo])";

  const char* src =
    "use \"pkg\"\n"
    "use @foo[None](x: USize)\n"

    "actor Main\n"
    "  new create(env: Env) =>\n"
    "    @foo(0)\n"
    "    Foo()";

  add_package("pkg", pkg);

  TEST_COMPILE(src);
}


TEST_F(CodegenFFITest, RuntimeCall)
{
  const char* src =
    "actor Main\n"
    "  new create(env: Env) =>\n"
    "    @pony_ctx[Pointer[None]]()";

  TEST_COMPILE(src);
}


TEST_F(CodegenFFITest, RuntimeCallConflict)
{
  const char* src =
    "actor Main\n"
    "  new create(env: Env) =>\n"
    "    @pony_ctx[None]()";

  TEST_ERRORS_1(src, "conflicting declarations for FFI function: return values "
    "have incompatible types");
}


TEST_F(CodegenFFITest, DeclParameterCountConflict)
{
  const char* pkg =
    "use @foo[None](x: U32)\n"

    "primitive Foo\n"
    "  fun apply() => @foo(0)";

  const char* src =
    "use \"pkg\"\n"
    "use @foo[None](x: U32, y: U32)\n"

    "actor Main\n"
    "  new create(env: Env) =>\n"
    "    @foo(0, 0)\n"
    "    Foo()";

  add_package("pkg", pkg);

  TEST_ERRORS_1(src, "conflicting declarations for FFI function: declarations "
    "have an incompatible number of parameters");
}


TEST_F(CodegenFFITest, DeclParameterTypeConflict)
{
  const char* pkg =
    "use @foo[None](x: U32)\n"

    "primitive Foo\n"
    "  fun apply() => @foo(0)";

  const char* src =
    "use \"pkg\"\n"
    "use @foo[None](x: U64)\n"

    "actor Main\n"
    "  new create(env: Env) =>\n"
    "    @foo(0)\n"
    "    Foo()";

  add_package("pkg", pkg);

  TEST_ERRORS_1(src, "conflicting declarations for FFI function: parameters "
    "have incompatible types");
}


TEST_F(CodegenFFITest, DeclReturnTypeConflict)
{
  const char* pkg =
    "use @foo[U32]()\n"

    "primitive Foo\n"
    "  fun apply() => @foo()";

  const char* src =
    "use \"pkg\"\n"
    "use @foo[U64]()\n"

    "actor Main\n"
    "  new create(env: Env) =>\n"
    "    @foo()\n"
    "    Foo()";

  add_package("pkg", pkg);

  TEST_ERRORS_1(src, "conflicting declarations for FFI function: return values "
    "have incompatible types");
}


TEST_F(CodegenFFITest, DeclReturnTypeVoidConflict)
{
  const char* pkg =
    "use @foo[None]()\n"

    "primitive Foo\n"
    "  fun apply() => @foo()";

  const char* src =
    "use \"pkg\"\n"
    "use @foo[Main]()\n"

    "actor Main\n"
    "  new create(env: Env) =>\n"
    "    @foo()\n"
    "    Foo()";

  add_package("pkg", pkg);

  TEST_ERRORS_1(src, "conflicting declarations for FFI function: return values "
    "have incompatible types");
}


TEST_F(CodegenFFITest, InternalABIFunctionConflict)
{
  const char* src =
    "actor Main\n"
    "  new create(env: Env) =>\n"
    "    @Main_Dispatch[None]()";

  TEST_ERRORS_1(src, "cannot use 'Main_Dispatch' as an FFI name: name is "
    "already in use by the internal ABI");
}


TEST_F(CodegenFFITest, InternalABIObjectConflict)
{
  const char* src =
    "actor Main\n"
    "  new create(env: Env) =>\n"
    "    @Main_Desc[None]()";

  TEST_ERRORS_1(src, "cannot use 'Main_Desc' as an FFI name: name is already "
    "in use by the internal ABI");
}
