#include <gtest/gtest.h>
#include <platform.h>

#include "util.h"

#define TEST_COMPILE(src) DO(test_compile(src, "ir"))


class CodegenIdentityTest : public PassTest
{};


extern "C"
{

EXPORT_SYMBOL uintptr_t ptr_to_int(void* p)
{
  return (uintptr_t)p;
}

}


TEST_F(CodegenIdentityTest, ObjectIsObject)
{
  const char* src =
    "class C1\n"
    "class C2\n"

    "actor Main\n"
    "  new create(env: Env) =>\n"
    "    let c1: Any = C1\n"
    "    let c2: Any = C2\n"
    "    if (c1 is c1) and (c1 isnt c2) then\n"
    "      @pony_exitcode[None](I32(1))\n"
    "    end";

  TEST_COMPILE(src);

  int exit_code = 0;
  ASSERT_TRUE(run_program(&exit_code));
  ASSERT_EQ(exit_code, 1);
}


TEST_F(CodegenIdentityTest, NumericIsNumeric)
{
  const char* src =
    "actor Main\n"
    "  new create(env: Env) =>\n"
    "    if (U32(0) is U32(0)) and (U32(0) isnt U32(1)) and\n"
    "      (U32(0) isnt U64(0))\n"
    "    then\n"
    "      @pony_exitcode[None](I32(1))\n"
    "    end";

  TEST_COMPILE(src);

  int exit_code = 0;
  ASSERT_TRUE(run_program(&exit_code));
  ASSERT_EQ(exit_code, 1);
}


TEST_F(CodegenIdentityTest, TupleIsTuple)
{
  const char* src =
    "actor Main\n"
    "  new create(env: Env) =>\n"
    "    let a: (Any, Any) = (U32(0), env)\n"
    "    let b: (Any, Any) = (U32(1), this)\n"
    "    if (a is a) and (a isnt b) then\n"
    "      @pony_exitcode[None](I32(1))\n"
    "    end";

  TEST_COMPILE(src);

  int exit_code = 0;
  ASSERT_TRUE(run_program(&exit_code));
  ASSERT_EQ(exit_code, 1);
}


TEST_F(CodegenIdentityTest, NumericIsBoxedNumeric)
{
  const char* src =
    "actor Main\n"
    "  new create(env: Env) =>\n"
    "    let boxed: (Any | (U32, U32)) = U32(0)\n"
    "    if (U32(0) is boxed) and (U32(1) isnt boxed) and\n"
    "      (U64(0) isnt boxed)\n"
    "    then\n"
    "      @pony_exitcode[None](I32(1))\n"
    "    end";

  TEST_COMPILE(src);

  int exit_code = 0;
  ASSERT_TRUE(run_program(&exit_code));
  ASSERT_EQ(exit_code, 1);
}


TEST_F(CodegenIdentityTest, BoxedNumericIsBoxedNumeric)
{
  const char* src =
    "actor Main\n"
    "  new create(env: Env) =>\n"
    "    let u32_0_a: (Any | (U32, U32)) = U32(0)\n"
    "    let u32_0_b: (Any | (U32, U32)) = U32(0)\n"
    "    let u32_1: (Any | (U32, U32)) = U32(1)\n"
    "    let u64_0: (Any | (U32, U32)) = U64(0)\n"
    "    if (u32_0_a is u32_0_a) and (u32_0_a is u32_0_b) and\n"
    "      (u32_0_a isnt u32_1) and (u32_0_a isnt u64_0)\n"
    "    then\n"
    "      @pony_exitcode[None](I32(1))\n"
    "    end";

  TEST_COMPILE(src);

  int exit_code = 0;
  ASSERT_TRUE(run_program(&exit_code));
  ASSERT_EQ(exit_code, 1);
}


TEST_F(CodegenIdentityTest, TupleIsBoxedTuple)
{
  const char* src =
    "actor Main\n"
    "  new create(env: Env) =>\n"
    "    let boxed: (Any | (U32, U32) | (U64, U64)) = (U32(0), U32(0))\n"
    "    if ((U32(0), U32(0)) is boxed) and ((U32(1), U32(0)) isnt boxed) and\n"
    "      ((U64(0), U64(0)) isnt boxed)\n"
    "    then\n"
    "      @pony_exitcode[None](I32(1))\n"
    "    end";

  TEST_COMPILE(src);

  int exit_code = 0;
  ASSERT_TRUE(run_program(&exit_code));
  ASSERT_EQ(exit_code, 1);
}


TEST_F(CodegenIdentityTest, BoxedTupleIsBoxedTuple)
{
  const char* src =
    "actor Main\n"
    "  new create(env: Env) =>\n"
    "    let t1_a: (Any | (U32, U32) | (U64, U64)) = (U32(0), U32(0))\n"
    "    let t1_b: (Any | (U32, U32) | (U64, U64)) = (U32(0), U32(0))\n"
    "    let t2: (Any | (U32, U32) | (U64, U64)) = (U32(1), U32(0))\n"
    "    let t3: (Any | (U32, U32) | (U64, U64)) = (U64(0), U64(0))\n"
    "    if (t1_a is t1_a) and (t1_a is t1_b) and (t1_a isnt t2) and\n"
    "      (t1_a isnt t3)\n"
    "    then\n"
    "      @pony_exitcode[None](I32(1))\n"
    "    end";

  TEST_COMPILE(src);

  int exit_code = 0;
  ASSERT_TRUE(run_program(&exit_code));
  ASSERT_EQ(exit_code, 1);
}


TEST_F(CodegenIdentityTest, TupleCardinality)
{
  const char* src =
    "actor Main\n"
    "  new create(env: Env) =>\n"
    "    if (env, this, this) isnt (env, this) then\n"
    "      @pony_exitcode[None](I32(1))\n"
    "    end";

  TEST_COMPILE(src);

  int exit_code = 0;
  ASSERT_TRUE(run_program(&exit_code));
  ASSERT_EQ(exit_code, 1);
}


TEST_F(CodegenIdentityTest, AbstractTypeNoSubtyping)
{
  const char* src =
    "actor Main\n"
    "  new create(env: Env) =>\n"
    "    let a: (Env | None) = None\n"
    "    let b: (Main | None) = None\n"
    "    if a is b then\n"
    "      @pony_exitcode[None](I32(1))\n"
    "    end";

  TEST_COMPILE(src);

  int exit_code = 0;
  ASSERT_TRUE(run_program(&exit_code));
  ASSERT_EQ(exit_code, 1);
}


TEST_F(CodegenIdentityTest, NestedTuple)
{
  const char* src =
    "actor Main\n"
    "  new create(env: Env) =>\n"
    "    let a: (((U8, U16) | None, U8) | None) = None\n"
    "    let b: (((U8, U16) | None, U8) | None) = None\n"
    "    if a is b then\n"
    "      @pony_exitcode[None](I32(1))\n"
    "    end";

  TEST_COMPILE(src);

  int exit_code = 0;
  ASSERT_TRUE(run_program(&exit_code));
  ASSERT_EQ(exit_code, 1);
}


TEST_F(CodegenIdentityTest, TupleDifferentTypes)
{
  const char* src =
    "actor Main\n"
    "  new create(env: Env) =>\n"
    "    let a: ((Any, U8) | None) = (U8(0), 0)\n"
    "    let b: ((U8, U8) | None) = (0, 0)\n"
    "    if a is b then \n"
    "      @pony_exitcode[None](I32(1))\n"
    "    end";

  TEST_COMPILE(src);

  int exit_code = 0;
  ASSERT_TRUE(run_program(&exit_code));
  ASSERT_EQ(exit_code, 1);
}


TEST_F(CodegenIdentityTest, DigestofObject)
{
  const char* src =
    "actor Main\n"
    "  new create(env: Env) =>\n"
    "    let dg = digestof env\n"
    "    if dg == @ptr_to_int[USize](env) then\n"
    "      @pony_exitcode[None](I32(1))\n"
    "    end";

  TEST_COMPILE(src);

  int exit_code = 0;
  ASSERT_TRUE(run_program(&exit_code));
  ASSERT_EQ(exit_code, 1);
}


TEST_F(CodegenIdentityTest, DigestofNumeric)
{
  const char* src =
    "actor Main\n"
    "  new create(env: Env) =>\n"
    "    let value = U32(5)\n"
    "    let dg = digestof value\n"
    "    if dg == 5 then\n"
    "      @pony_exitcode[None](I32(1))\n"
    "    end";

  TEST_COMPILE(src);

  int exit_code = 0;
  ASSERT_TRUE(run_program(&exit_code));
  ASSERT_EQ(exit_code, 1);
}


TEST_F(CodegenIdentityTest, DigestofTuple)
{
  const char* src =
    "actor Main\n"
    "  new create(env: Env) =>\n"
    "    let value = (this, env)\n"
    "    let dg = digestof value\n"
    "    if dg == ((digestof this) xor (digestof env)) then\n"
    "      @pony_exitcode[None](I32(1))\n"
    "    end";

  TEST_COMPILE(src);

  int exit_code = 0;
  ASSERT_TRUE(run_program(&exit_code));
  ASSERT_EQ(exit_code, 1);
}


TEST_F(CodegenIdentityTest, DigestofBoxedNumeric)
{
  const char* src =
    "actor Main\n"
    "  new create(env: Env) =>\n"
    "    let boxed: (Any | (U32, U32)) = U32(5)\n"
    "    let dg = digestof boxed\n"
    "    if dg == 5 then\n"
    "      @pony_exitcode[None](I32(1))\n"
    "    end";

  TEST_COMPILE(src);

  int exit_code = 0;
  ASSERT_TRUE(run_program(&exit_code));
  ASSERT_EQ(exit_code, 1);
}


TEST_F(CodegenIdentityTest, DigestofBoxedTuple)
{
  const char* src =
    "actor Main\n"
    "  new create(env: Env) =>\n"
    "    let boxed: (Any | (Main, Env)) = (this, env)\n"
    "    let dg = digestof boxed\n"
    "    if dg == ((digestof this) xor (digestof env)) then\n"
    "      @pony_exitcode[None](I32(1))\n"
    "    end";

  TEST_COMPILE(src);

  int exit_code = 0;
  ASSERT_TRUE(run_program(&exit_code));
  ASSERT_EQ(exit_code, 1);
}
